/*
btupdt.c Version 0.1.0  Btree update record
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* return code 0 is good update */
/* return code 1 is record not found */
/* return code 3 is lock pending prior to update */
/* return code 4 is lock error */

#include <bthash.h>

int btupdt(bthdr,ofst,rcd)
btreefmt *bthdr;
off_t ofst;
unsigned char *rcd;
   {
   int sigrslt;
   int btgrslt;
   int lockrslt;
   unsigned char *p,*q,*r;
   unsigned char *buf;
   bthdr->reason = 0;
   sigrslt = strcmp(bthdr->signature,BT_SIGNATURE);
   if (sigrslt)
      {
      fprintf(stderr,"btisrt: non-btree "
	 "update\n");
      bthdr->reason = BTREE_ERR + 2206;
      return(BTREE_ERR + 2206);
      } /* not a btree header */
   p = (unsigned char *) rcd + bthdr->dletofst;
   if (*p == DELETED)
      {
      fprintf(stderr,"btupdt: updating "
	 "record has delete byte set\n");
      fprintf(stderr,"btupdt: call "
	 "btdlet for delete\n");
      bthdr->reason = BTREE_ERR + 2205; 
      return(BTREE_ERR + 2205); 
      } /* delete byte set */
   lockrslt = lockf(bthdr->hndl,F_TEST,0);
   if (lockrslt < 0)
      {
      return(UPDTLOCK3);
      } /* can't update when file is locked */
   lockrslt = lockf(bthdr->hndl,F_TLOCK,0);
   if (lockrslt < 0)
      {
      fprintf(stderr,"btupdt: lock error on %s\n",
	 bthdr->dbname);
      perror("Lock error");
      return(UPDTLOCK4);
      } /* can't update if lock error */
   if (ofst == ROOT
      && bthdr->head->next != bthdr->tail)
      {
      fprintf(stderr,"btupdt: inserting at root "
	 "with non-empty stack\n");
      return(BTREE_ERR + 2201);
      } /* insert loop */
   buf = (unsigned char *) malloc(bthdr->rcdlen + 16);
   if (buf == NULL)
      {
      fprintf(stderr,"btupdt: out of memory\n");
      return(BTREE_ERR + 2202);
      } /* out of mem */
   p = (unsigned char *) rcd + bthdr->keyofst;
   btgrslt = btget(bthdr,ofst,p,buf);
   free(buf);
   if (btgrslt == FRSTPAPA) return(UPDTERR);
   else if (btgrslt) return(btgrslt);
   p = (unsigned char *) bthdr->blk + bthdr->blkhdr
      + (bthdr->rcdnum * bthdr->rcdnode)
      - bthdr->rcdnode + bthdr->rcdhdr;
   q = (unsigned char *) p + bthdr->rcdlen;
   r = (unsigned char *) rcd;
   if (*(p + bthdr->dletofst) == DELETED)
      {
      fprintf(stderr,"btupdt: matching "
	 "record has delete byte set\n");
      bthdr->reason = BTREE_ERR + 2204;
      return(BTREE_ERR + 2204);
      } /* deleting twice */
   while (p < q) *p++ = *r++;
   btputblk(bthdr,bthdr->ofst,bthdr->blk);
   if (bthdr->reason) return(BTREE_ERR + 2203);
   p = (unsigned char *) bthdr->blk;
   q = (unsigned char *) bthdr->blk + bthdr->blksz;
   while (p < q) *p++ = '\0';
   bthdr->ofst = (off_t) 0;
   if (bthdr->lock != LOCK)
      {
      lockrslt = lockf(bthdr->hndl,F_ULOCK,0);
      if (lockrslt < 0)
         {
         fprintf(stderr,"btupdt: error unlocking "
            "%s after insertion\n",
            bthdr->dbname);
         perror("Unlock error");
         bthdr->reason  = SYS_ERR + errno;
	 return(SYS_ERR + errno);
         }
      } /* if no lock during open, unlock */
   return(UPDTGOOD);
   } /* btupdt */
